// Last Updated: 13-06-20
// GPS Tracking System V2 - ESP32 DEV Board
// System sends email to defined Gmail account
// Remember to remove the RX pin, before downloading a new program
// Remember when opening a SD file the system will overwrite anything in that
// file as it starts from the beginnin of the file, no append data option available.
// Gmail shows a GPX file as a Binary file, whereas a KML file has a "</>" symbol in front

#include <SD.h>
#include <sd_defines.h>
#include <sd_diskio.h>

#include "ESP32_MailClient.h"
#include "SPIFFS.h"
#include <WiFiClient.h>
WiFiClient client;
#include <math.h>;
float calculate_lat,calculate_long,result,current_latitude,current_longitude;
float last_latitude,last_longitude,save_latitude,save_longitude;
float start_latitude,start_longitude,distance_between_points,working_result,working_result2,distance_between_current_points;
float get_latitude,get_longitude,calculated_speed,calculated_distance,constant,latest_longitude,latest_latitude;
int test,current_seconds,save_seconds,period;
    
const char* ssid = "yourRouterID";
const char* password = "yourRouterpassword";

// To send Email using Gmail use port 465 (SSL) and SMTP Server smtp.gmail.com
// YOU MUST ENABLE less secure app option https://myaccount.google.com/lesssecureapps?pli=1
#define emailSenderAccount    "youraccount@gmail.com"    
#define emailSenderPassword   "yourpassword"
#define emailRecipient        "accounttosendto@gmail.com"
#define smtpServer            "smtp.gmail.com"
#define smtpServerPort        465
#define emailSubject          "GPS Route Test"

// The Email Sending data object contains config and data to send, can use 587
SMTPData smtpData;

// Callback function to get the Email sending status
void sendCallback(SendStatus info);

#include <TinyGPS++.h>           // Include TinyGPS++ library
#include <SoftwareSerial.h>      // Include software serial library

#include "RTClib.h" 
RTC_DS3231 rtc;
int which_one;
int route_count,button_pressed,counter,degrees_whole,degrees_minutes,degrees_seconds;
int connection_counter = 0;
String fileID,hours,minutes,seconds,years,months,days;
char recording,file_format;
String directions[17] = {"N","NNE","NE","ENE","E","ESE","SE","SSE","S","SSW","SW","WSW","W","WNW","NW","NNW","N"};
int direction_in_degrees;
String compass_course;
char file_transfer_required,first_recording;
String filename,extension;
int filesarrayCounter,i,found_any_files;
String files[10];

#include <SPI.h>
#include <SD.h>
const int chipSelect = 5; //4
File root;
File dataFile;

// Screen setup
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>


#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET     4 // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

#include <Wire.h>

TinyGPSPlus gps; 
static const int RXPin = 3, TXPin = 1;  //RXPin = D4, TXPin = D3;
static const uint32_t GPSBaud = 9600;
SoftwareSerial SoftSerial(RXPin,TXPin);    // Configure SoftSerial library

void setup(void) 
{
  rtc.begin(); // Initialize the rtc object
//  rtc.adjust(DateTime(2020, 6, 9, 13, 49, 00));

  // setup 34 to receive push button inputs
  pinMode(15, INPUT);
  which_one = 1;
  recording = 'N';
  file_transfer_required = 'N';
  
 // Init 0.96" LCD screen
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.clearDisplay();  
  display.setCursor(10,15);
  display.print("GPS Tracking V2");
  display.display(); 
  delay(2000);  
  display.clearDisplay(); 
    
//Start softSerial at this point so as to not  
   SoftSerial.begin(9600); 
   
//   sd Card Initialization
  pinMode(SS, OUTPUT);
  if (SD.begin(chipSelect))
  {    
    display.setCursor(10,10);
    display.print("Card Initalised ");
    display.display();
    delay(2000);
    display.setCursor(10,20);
    root = SD.open("/");
    printDirectory(root);
    if (found_any_files == 0)
    {   display.clearDisplay();    
        display.setCursor(10,10); 
        display.print("No files found  ");
        display.display();
    }    
  } else 
  {
    display.clearDisplay(); 
    display.setCursor(10,10);
    display.print("Card failed ");
    display.display();
  }
  delay(2000);

 // Ask user if file transfer is required if files are available
 if (found_any_files != 0)
 {
    file_transfer_required = 'N';
    display.clearDisplay();    
    display.setCursor(10,20); 
    display.print("E-Mail Send mode");
    display.setCursor(10,30); 
    display.print("press button to");
    display.setCursor(10,40); 
    display.print("enable");
    display.display();
    counter = 1;
    while (counter <= 50)
    {
      button_pressed = digitalRead(15);
      if (button_pressed == HIGH)
      {
         display.clearDisplay();
         display.setCursor(10,10); 
         display.print("E-Mail sending ");
         display.setCursor(10,20); 
         display.print("mode enabled   ");
         display.display();
         delay(2000);
         file_transfer_required = 'Y';
         display.clearDisplay();
         break;
      }
      counter = counter + 1;  
      delay(100);
    }      
    
    // If E-mail required then attempt to connect to local WiFi otherwise do not connect
    if (file_transfer_required == 'Y')
    {
        WiFi.begin(ssid, password);
        display.clearDisplay();
        display.setCursor(10,20);
        display.print("Connecting");
        display.setCursor(10,30);
        display.display();
        // Wait for connection
        while (WiFi.status() != WL_CONNECTED) 
        {
          if (connection_counter <= 10)
          {
             display.print(".");
             display.display();
          } else
          {
             display.setCursor(10,30);      
             display.print("Waiting.........");
             display.display();
          }
          delay(500);
          connection_counter = connection_counter +1;
          if (connection_counter > 50)
          {
            break;
          }
        }
        display.clearDisplay();
        display.setCursor(0,10);
        if ( WiFi.status() == WL_CONNECTED) 
        {
            display.println(ssid);
            display.setCursor(0,30);
            display.print("IP:");
            display.print(WiFi.localIP());
            display.display();
            delay(2000);
            // Files available on SD-Card, WiFi connected, E-mail requested so
            perform_email_send();
        } else 
        {
            display.clearDisplay();
            display.setCursor(10,10); 
            display.print("WiFi failed    ");
            display.display();    
        }
    }
 } 
    delay(2000);
    display.clearDisplay();   
  
// File transfer has been completed or not required, all files removed from SD-card 
// so ask user for recording format.
 file_format = 'G'; // Default to GPX mode  
        display.clearDisplay();
        display.setCursor(10,10); 
        display.print("File Format     ");
        display.setCursor(10,20); 
        display.print("defaults to GPX ");
        display.display();        
        delay(2000); 
        display.clearDisplay();    
        display.setCursor(10,10); 
        display.print("OR Press button ");
        display.setCursor(10,20); 
        display.print("for KML format  ");
        display.display();        
        counter = 1;
        while (counter <= 50)
        {
          button_pressed = digitalRead(15);
          if (button_pressed == HIGH)
          {
             display.clearDisplay();
             display.setCursor(10,10); 
             display.print("KML file format");
             display.setCursor(10,20); 
             display.print("selected       ");
             display.display();             
             delay(2000);
             file_format = 'K';
             display.clearDisplay();
             break;
          }
          counter = counter + 1;  
          delay(100);
        }
        if (counter >= 50)
        {
             display.clearDisplay();
             display.setCursor(10,10); 
             display.print("GPX file format");
             display.setCursor(10,20); 
             display.print("selected       ");
             display.display();             
             delay(2000);   
        }
        display.clearDisplay();
        display.setCursor(10,10); 
        display.print("Waiting for GPS ");
        display.setCursor(10,20); 
        display.print("signal......... ");
        display.display();
   save_latitude = 0;
   save_longitude = 0;        
}

void loop() 
{
   
  while (SoftSerial.available() > 0) 
    {
      button_pressed =0;
      if (gps.encode(SoftSerial.read())) 
      {
        if (gps.location.isUpdated()) //isValid
        {
            button_pressed = digitalRead(15);
            if (button_pressed == HIGH and which_one == 1) // pin high -> pressed 15
            { 
              display.clearDisplay();
              start_file();
              display.setCursor(10,10); 
              display.print("Route Monitoring");
              display.setCursor(10,20); 
              display.print("File:"); 
              display.print(fileID);
              display.print(" open");  
              display.display();            
              which_one = 2;
              distance_between_points = 0;
              calculated_speed = 0;
              route_count = 1;
              button_pressed = 0;
              recording = 'Y';
              calculated_distance = 0;
              delay(2000);
            }  
            if (button_pressed == HIGH and which_one == 2) // pin High -> pressed    
            {    
              display.clearDisplay();     
              display.setCursor(10,10); 
              display.print("Route Monitoring");
              display.setCursor(10,20); 
              display.print("has been stopped");
              display.display();          
              finish_file();
              which_one = 1;
              recording = 'N';
              button_pressed = 0;
              delay(2000);
            }  
            current_latitude = gps.location.lat();
            current_longitude = gps.location.lng();
            get_date_and_time(); 
            if (save_latitude == 0 and save_longitude == 0 )
            {
              save_latitude = current_latitude;
              save_longitude = current_longitude;
              save_seconds = current_seconds;
            }
            display.clearDisplay(); 
            display.setCursor(20,10); 
            display.print("Latitude:");
            display.setCursor(20,20); 
            display_latitude(current_latitude);
            display.setCursor(20,30); 
            display.print("Longitude:");
            display.setCursor(20,40);           
            display_longitude(current_longitude);                       
            display.display(); 
            delay(2000);
            display.clearDisplay();          
            display.setCursor(10,10); 
            display.print("Speed: ");
            calculate_speed_and_distance();            
            display.print(calculated_speed,0);
            display.print(" Km/h");
            display.setCursor(10,20);           
            display.print("Course: "); 
            get_course(gps.course.deg());
            display.print(" No:");
            display.print(gps.satellites.value());
            display.setCursor(10,30); 
            display.print("Altitude: ");
            display.print(gps.altitude.meters(),0);
            display.print("m");
            display.setCursor(10,40);
            display.print("Time: ");
            display.print(hours);
            display.print(":");     
            display.print(minutes);
            display.print(":");     
            display.print(seconds);
            display.setCursor(10,50);
            display.print("Date: ");            
            display.print(days);
            display.print("-");
            display.print(months);
            display.print("-");           
            display.print(years);           
            display.display();          
            delay(2000);
            if (recording == 'Y') // Route Recording started
            {
              write_sd_card(); // Write entry to sd-Card
              display.clearDisplay();
              display.setCursor(10,10); 
              display.print("Track: ");
              display.print(fileID);
              display.print(".");
              if (file_format == 'G')
              {
                display.print("GPX");
              }else 
              {
                display.print("KML");
              }
              display.setCursor(10,30); 
              display.print("WayPoints: ");
              display.print(route_count);  
              display.setCursor(10,50); 
              display.print("Distance: ");
              display.print(calculated_distance,2);
              display.print(" Km");
              display.display();
              route_count = route_count + 1;  
              display.display();                
              delay(2000);
            }            
        } else {
            display.clearDisplay();
            display.setCursor(10,10); 
            display.print("Waiting for GPS ");
            display.setCursor(10,20); 
            display.print("signal......... ");
            display.display();
        }
      } // end softserial
    } // end while      
}

void calculate_speed_and_distance()
{
  constant = PI / 180;

  // test only
//    current_latitude = -43.443738;
//    current_longitude = 172.632913; 
//    save_latitude = -43.443295;
//    save_longitude = 172.630387; 
//    // distance 0.2098km
//    current_seconds = 40;
//    save_seconds = 30;  
//    // speed = 75.53km/h  
                  
  // we know the distance travelled over a period of 2 seconds
  // so simply multiply value by 1800 to get km/h 
    // get radians value of current latitude gps reading
    latest_latitude = current_latitude * constant;
    // get radians value of current longitude gps reading
    latest_longitude = current_longitude * constant; 
    // get radians value of current latitude gps reading
    get_latitude = save_latitude * constant;
    // get radians value of current longitude gps reading
    get_longitude = save_longitude * constant;    
    // Haversine Formula - Calculate geographic distance on earth
    // between two points: current_latitude and save_latitude, 
    // current_longitude and save_longitude
    calculate_lat = latest_latitude - get_latitude;
    calculate_long = latest_longitude - get_longitude;
    // Sin of latitude divided by 2 squared + 
    // Sin of longitude divided by 2 squared all multipled by 
    // Cos of first latitude multiplied by the Cos of second latitude.
    working_result = (sin(calculate_lat/2) * sin(calculate_lat/2)) +
    (sin(calculate_long/2) * sin(calculate_long/2)) *
    (cos(get_latitude) * cos(latest_latitude));
    working_result2 = 2 * atan2(sqrt(working_result),sqrt(1-working_result)); 
    // Distance in Km divided by time.
    distance_between_points = 6371 * working_result2; // in kilometers
    // Period of time between one update and the next
                                                 
    if(save_seconds > current_seconds)
    {
      period = (60 - save_seconds) + current_seconds;
    }else
    {
      period = current_seconds - save_seconds;
    }
    if (period > 0 and distance_between_points > 0)
    {
         calculated_speed = distance_between_points * (3600 / period);
         // If speed value is very small less than 5 then set to 0
         if (calculated_speed <= 5 )
         {
          calculated_speed = 0;
         }
    } else
    {
      calculated_speed = 0;
    }
    if (calculated_speed != 0)
    {
        calculated_distance =  calculated_distance + distance_between_points;  // distance in Km
    }
    // Save old position
    save_latitude = current_latitude;
    save_longitude = current_longitude;
    save_seconds = current_seconds;  
}

void start_file()
{
    DateTime now = rtc.now();
    hours = String(now.hour());
    minutes = String(now.minute());
    seconds = String(now.second());
    current_seconds = now.second();    
    if (hours.length() == 1)
    {
       hours = "0" + hours;     
    }
    if (minutes.length() == 1)
    {
       minutes = "0" + minutes;     
    }

    if (seconds.length() == 1)
    {
       seconds = "0" + seconds;     
    }        
  fileID = hours+minutes+seconds;
  if (file_format == 'G')
  {      // Write out GPX file header
      dataFile = SD.open("/"+fileID+".gpx", FILE_WRITE);
      dataFile.println("<?xml version=\"1.0\"?>");
      dataFile.println("<gpx>");
      dataFile.println("<trk>");
      dataFile.println("<name>"+fileID+" GPS Route</name>");
      dataFile.println("<extensions>");
      dataFile.println("<gpxx:TrackExtension xmlns:gpxx=\"http://www.garmin.com/xmlschemas/GpxExtensions/v3\">");
      dataFile.println("<gpxx:DisplayColor>Black</gpxx:DisplayColor>");
      dataFile.println("</gpxx:TrackExtension>");
      dataFile.println("</extensions>");
      dataFile.println("<trkseg>");  
  }else 
  {     // Write out KML file header
     dataFile = SD.open("/"+fileID+".kml", FILE_WRITE);
     dataFile.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
     dataFile.println("<kml xmlns=\"http://www.opengis.net/kml/2.2\">");
     dataFile.println("<Document>");
     dataFile.println("<name>GPS Route</name>");
     dataFile.println("<Placemark>");
     dataFile.println("<name>"+fileID+" GPS Route</name>");
     dataFile.println("<LineString>");
     dataFile.println("<coordinates>");  
  }
}

void write_sd_card()
{
    // Write data out to file on sd-Card
      if(file_format == 'G')
      {    // GPX file format
        dataFile.print("<trkpt lat=\"");
        dataFile.print(gps.location.lat(),6);  
        dataFile.print("\" lon=\"");
        dataFile.print(gps.location.lng(),6);
        dataFile.println("\">");
        dataFile.print("<ele>");
        dataFile.print(gps.altitude.meters());
        dataFile.println("</ele>");
        dataFile.print("<time>");
        get_date_and_time();   
        dataFile.print(years+"-"+months+"-"+days+"T");
        dataFile.print(hours+":"+minutes+":"+seconds+"Z");
        dataFile.println("</time>");
        dataFile.println("</trkpt>");
      } else // KML file format
      {
        dataFile.print(gps.location.lng(),6);
        dataFile.print(",");
        dataFile.print(gps.location.lat(),6);  
        dataFile.println(",2357");          
      }      
}

void finish_file()
{
    if (file_format == 'G')
    {  // Write out file end data in GPX format
      dataFile.println("</trkseg>");  
      dataFile.println("</trk>");
      dataFile.println("</gpx>");
    }else 
    {  // Write out file end data in KML format
      dataFile.println("</coordinates>");
      dataFile.println("</LineString>");
      dataFile.println("<Style>");
      dataFile.println("<LineStyle>");
      dataFile.println("<color>#ff000000</color>");
      dataFile.println("<width>5</width>");
      dataFile.println("</LineStyle>");
      dataFile.println("</Style>");
      dataFile.println("</Placemark>");
      dataFile.println("</Document>");
      dataFile.println("</kml>");          
    }
    dataFile.close();  
}

void get_date_and_time()
{
    DateTime now = rtc.now();
        hours = String(now.hour());
        minutes = String(now.minute());
        seconds = String(now.second());
        current_seconds = now.second();
        years = String(now.year());
        months = String(now.month());
        days = String(now.day());      
        if (hours.length() == 1)
        {
           hours = "0" + hours;     
        }
        if (minutes.length() == 1)
        {
           minutes = "0" + minutes;     
        }
        if (seconds.length() == 1)
        {
           seconds = "0" + seconds;     
        } 
        if (months.length() == 1)
        {
           months = "0" + months;     
        }  
        if (days.length() == 1)
        {
           days = "0" + days;     
        }                       
}

void display_latitude(float degrees_full)
{
  degrees_whole = int(degrees_full) * -1;  // negative latitude converted to position for calc only
  int degrees_whole_display = int(degrees_full) * -1;
  float minutesRemainder = ((degrees_full*-1) - degrees_whole) * 60;
  int arcMinutes = minutesRemainder;
  int arcSeconds = (minutesRemainder - arcMinutes) * 60;
  display.print(degrees_whole_display);
  display.print((char)248);
  display.print(arcMinutes);
  display.print("'");
  display.print(arcSeconds);
  if (degrees_full < 0)
  {
      display.print("\"S");
  } else
  {
      display.print("\"N");        
  }
  display.display();
}

void display_longitude(float degrees_full)
{
  degrees_whole = int(degrees_full);
  float minutesRemainder = (degrees_full - degrees_whole)*60;
  int arcMinutes = minutesRemainder;
  int arcSeconds = (minutesRemainder - arcMinutes) * 60;
  display.print(degrees_whole);
  display.print((char)248); //degrees symbol
  display.print(arcMinutes);
  display.print("'");
  display.print(arcSeconds);
  if (degrees_full < 0)
  {
      display.print("\"W");
  } else
  {
      display.print("\"E");        
  } 
  display.display();  
}
void get_course(float gps_course)
{
  direction_in_degrees = int(gps_course/22.5)+1;
  compass_course = directions[direction_in_degrees];
  display.print(compass_course);
  display.display();  
}

void printDirectory(File dir)
{
      found_any_files = 0;
      filesarrayCounter = 0;
      dir.rewindDirectory();
      while(true)
      {
          File entry =  dir.openNextFile();
          if (! entry) 
          {
               // no more files
               // return to the first file in the directory
               dir.rewindDirectory();
           break;
         }    
        filename = entry.name();
        extension = filename.substring(8,11);
        extension.toUpperCase();
        if (extension == "GPX" or extension == "KML")
        {
          files[filesarrayCounter] = filename.substring(1,11);  
          filesarrayCounter = filesarrayCounter + 1;
          found_any_files = 1;
          Serial.println(filesarrayCounter);
        }
      }
      for (int i=0; i<= filesarrayCounter-1; i++)
      {
          display.clearDisplay();    
          display.setCursor(10,10); 
          display.print("Files Found: ");
          display.setCursor(10,20); 
          display.print(files[i]);
          display.display();          
          delay(1000); 
          Serial.println(files[i]);
      }      
}
void perform_email_send()
{
   for ( i = 0; i <=filesarrayCounter-1; i++)
    {       
       get_date_and_time(); 
        display.clearDisplay();
        display.setCursor(10,10); 
        display.print("E-Mail file send");
        display.setCursor(10,20); 
        display.print("working.........");
        display.display();  
        // Set the SMTP Server Email host, port, account and password
        smtpData.setLogin(smtpServer, smtpServerPort, emailSenderAccount, emailSenderPassword);
      
        // For library version 1.2.0 and later which STARTTLS protocol was supported,the STARTTLS will be 
        // enabled automatically when port 587 was used, or enable it manually using setSTARTTLS function.
        //smtpData.setSTARTTLS(true);
      
        // Set the sender name and Email
        smtpData.setSender("ESP32", emailSenderAccount);
      
        // Set Email priority or importance High, Normal, Low or 1 to 5 (1 is highest)
        smtpData.setPriority("High");
      
        // Set the subject emailSubject
        smtpData.setSubject("GPS Route for "+files[i]);
      
        // Set the email message in text format (raw)
        smtpData.setMessage("Please find attached GPS Route file.", false);
      
        // Add recipients, you can add more than one recipient
        smtpData.addRecipient(emailRecipient);

        smtpData.addAttachFile("/"+files[i]);
      
        //Start sending Email, can be set callback function to track the status
        if (MailClient.sendMail(smtpData))
        {
          display.clearDisplay();
          display.setCursor(10,10); 
          display.print("E-Mail sent for "); 
          display.setCursor(10,20); 
          display.print("route:");
          display.print(files[i]);
          display.display();  
          // Sucessful send so remove file from SD card.        
          SD.remove("/"+files[i]); 
          delay(1000); 
        }else 
        {
          display.clearDisplay();
          display.setCursor(10,10); 
          display.print("E-Mail failed   ");
          display.display(); 
          delay(2000); 
        }
        //Clear all data from Email object to free memory
        smtpData.empty();
        delay(1000);
    }
    WiFi.disconnect();
}
